import os
from urllib.parse import urlparse
from typing import Dict, List
from pydantic import FileUrl
from ...models.config import QualityIndicator, QualityIndicatorType, CoverageBaseService, CoverageService, CoverageGeoJson, CoverageGeoPackage
from ...models.exceptions import DokAnalysisException


def get_coverage_indicator(quality_indicators: List[QualityIndicator]) -> QualityIndicator:
    if len(quality_indicators) == 0:
        return

    coverage_indicators = [
        indicator for indicator in quality_indicators if indicator.type == QualityIndicatorType.COVERAGE]

    if len(coverage_indicators) == 0:
        return
    elif len(coverage_indicators) > 1:
        raise DokAnalysisException(
            'A dataset can only have one coverage quality indicator')

    return coverage_indicators[0]


def get_coverage_service_config_data(coverage_indicator: QualityIndicator) -> Dict:
    coverage_svc: CoverageBaseService = None

    if coverage_indicator.wfs:
        coverage_svc = coverage_indicator.wfs
    elif coverage_indicator.arcgis:
        coverage_svc = coverage_indicator.arcgis
    elif coverage_indicator.geojson:
        coverage_svc = coverage_indicator.geojson
    elif coverage_indicator.gpkg:
        coverage_svc = coverage_indicator.gpkg

    if not coverage_svc:
        return None

    data = {
        'planning_guidance_id': coverage_svc.planning_guidance_id,
        'building_guidance_id': coverage_svc.building_guidance_id
    }

    if isinstance(coverage_svc, CoverageService):
        data['url'] = coverage_svc.url
        data['layer'] = coverage_svc.layer
    elif isinstance(coverage_svc, CoverageGeoJson) or isinstance(coverage_svc, CoverageGeoPackage):
        data['layer'] = coverage_svc.layer

        if isinstance(coverage_svc.url, FileUrl):
            parsed = urlparse(coverage_svc.url)
            data['url'] = os.path.basename(parsed.path)
        else:
            data['url'] = coverage_svc.url

    return data


__all__ = ['get_coverage_indicator', 'get_coverage_service_config_data']



import logging
from os import path
import json
from jsonschema import validate

_LOGGER = logging.getLogger(__name__)

_FILENAME = 'resources/no.geonorge.dokanalyse.v1.input.schema.json'
_DIR_PATH = path.dirname(path.realpath(__file__))
_FILE_PATH = path.abspath(path.join(_DIR_PATH, '../..', _FILENAME))


def request_is_valid(data) -> bool:
    with open(_FILE_PATH, 'r') as file:
        schema = json.load(file)

    try:
        validate(instance=data, schema=schema)
        return True
    except Exception as error:
        _LOGGER.error(str(error))
        return False


__all__ = ['request_is_valid']



import logging
from os import path, makedirs
from logging.handlers import TimedRotatingFileHandler
from .constants import APP_FILES_DIR
from .correlation_id_middleware import get_correlation_id


class CorrelationIdFilter(logging.Filter):
    def filter(self, record):
        record.correlation_id = get_correlation_id() or '-'
        return True


def setup() -> None:
    filename = path.join(APP_FILES_DIR, 'logs/dokanalyse.log')
    dirname = path.dirname(filename)

    if not path.exists(dirname):
        makedirs(dirname)

    handler = TimedRotatingFileHandler(
        filename, when='midnight', backupCount=30)

    handler.addFilter(CorrelationIdFilter())

    log_format = \
        '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - [ID: %(correlation_id)s] - %(message)s'

    handler.setFormatter(logging.Formatter(log_format))
    handler.namer = lambda name: name.replace('.log', '') + '.log'

    logging.root.setLevel(logging.INFO)
    logging.root.addHandler(handler)

    logger = logging.getLogger('azure')
    logger.setLevel(logging.WARNING)

__all__ = ['setup']


from pydantic import BaseModel, root_validator
from typing import Optional, Dict
from .quality_indicator_type import QualityIndicatorType
from . import CoverageService, CoverageGeoJson, CoverageGeoPackage


class QualityIndicator(BaseModel):
    type: QualityIndicatorType
    quality_dimension_id: str
    quality_dimension_name: str
    quality_warning_text: str
    warning_threshold: Optional[str]
    property: Optional[str] = None
    input_filter: Optional[str] = None
    wfs: Optional[CoverageService] = None
    arcgis: Optional[CoverageService] = None
    geojson: Optional[CoverageGeoJson] = None
    gpkg: Optional[CoverageGeoPackage] = None
    disabled: Optional[bool] = False

    @root_validator(pre=False)
    def check_coverage(cls, values: Dict) -> Dict:
        type = values.get('type')

        if type == QualityIndicatorType.COVERAGE and not 'wfs' in values and not 'arcgis' in values and not 'geojson' in values and not 'gpkg' in values:
            raise ValueError(
                'If the quality indicator type is "coverage", either the properties "wfs", "arcgis", "geojson" or "gpkg" must be set')

        return values



import logging
import time
import json
from uuid import UUID
from typing import List, Dict
from osgeo import ogr
from ..codelist import get_codelist
from ...drivers.ogc_api import query_ogc_api
from ...services.kartkatalog import get_kartkatalog_metadata
from ...utils.helpers.geometry import geometry_from_json
from ...utils.helpers.common import from_camel_case
from ...models.fact_part import FactPart

_LOGGER = logging.getLogger(__name__)

_METADATA_ID = UUID('900206a8-686f-4591-9394-327eb02d0899')
_LAYER_NAME = 'veglenke'
_API_BASE_URL = 'https://ogcapitest.kartverket.no/rest/services/forenklet_elveg_2_0/collections'
_TIMEOUT = 10


async def get_roads(geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int) -> FactPart:
    dataset = await get_kartkatalog_metadata(_METADATA_ID)
    data = await _get_data(geometry, epsg)

    return FactPart(geometry, epsg, orig_epsg, buffer, dataset, [f'intersect {_LAYER_NAME}'], data)


async def _get_data(geometry: ogr.Geometry, epsg: int) -> List[Dict]:
    start = time.time()
    status, response = await query_ogc_api(_API_BASE_URL, _LAYER_NAME, 'senterlinje', geometry, None, epsg, epsg, _TIMEOUT)
    end = time.time()

    if response is None:
        _LOGGER.error(
            f'Fact sheet: Could not get roads from Elveg OGC API (status {status})')
        return None

    # autopep8: off
    _LOGGER.info(f'Fact sheet: Got roads from Elveg OGC API: {round(end - start, 2)} sec.')
    # autopep8: on

    return await _map_response(response)


async def _map_response(response: Dict) -> List[Dict]:
    features: List[Dict] = response.get('features', [])
    road_categories = await get_codelist('vegkategori')
    road_types = {}

    for feature in features:
        json_str = json.dumps(feature['geometry'])
        geometry = geometry_from_json(json_str)
        props: Dict = feature.get('properties')
        road_type = props.get('typeVeg')

        if road_type == 'enkelBilveg':
            road_category = props.get('vegsystemreferanse', {}).get(
                'vegsystem', {}).get('vegkategori')

            if road_category is not None:
                entry = next(
                    (rc for rc in road_categories if rc['value'] == road_category), {})
                road_type = entry.get('label', road_type)

        if road_type in road_types:
            road_types[road_type] += geometry.Length()
        else:
            road_types[road_type] = geometry.Length()

    result: List[Dict] = []

    for key, value in road_types.items():
        result.append({
            'roadType': from_camel_case(key),
            'length': round(value, 2)
        })

    return result


__all__ = ['get_roads']



import logging
from typing import Tuple
import aiohttp
import asyncio
from ..models.map_image_payload import MapImagePayload
from ..utils.constants import MAP_IMAGE_API_URL

_LOGGER = logging.getLogger(__name__)
_TIMEOUT = 30


async def create_map_image(payload: MapImagePayload) -> Tuple[int, bytes]:
    data = payload.to_dict()

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(MAP_IMAGE_API_URL, json=data, timeout=_TIMEOUT) as response:
                if response.status != 200:
                    _LOGGER.error(
                        f'Could not generate map image (status {response.status})')
                    return response.status, None

                return 200, await response.read()
    except asyncio.TimeoutError:
        _LOGGER.error(f'Could not generate map image (status 408)')
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


__all__ = ['create_map_image']



import logging
import json
import os
from urllib.parse import urlparse
from typing import List, Dict, Tuple, Union
from pydantic import HttpUrl, FileUrl
from osgeo import ogr, osr
import aiohttp
import asyncio
from async_lru import alru_cache
from ..utils.helpers.geometry import create_feature_collection, transform_geometry

_LOGGER = logging.getLogger(__name__)
_CACHE_TTL = 86400


async def query_geojson(url: Union[HttpUrl, FileUrl], filter: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Dict:
    geojson = await _get_geojson(url, timeout)

    if not geojson:
        return None

    driver: ogr.Driver = ogr.GetDriverByName('GeoJSON')
    data_source: ogr.DataSource = driver.Open(geojson)
    layer: ogr.Layer = data_source.GetLayer(0)

    sr: osr.SpatialReference = layer.GetSpatialRef()
    auth_code: str = sr.GetAuthorityCode(None)
    geojson_epsg = int(auth_code)

    if geojson_epsg != epsg:
        input_geometry = transform_geometry(geometry, epsg, geojson_epsg)
    else:
        input_geometry = geometry

    layer.SetSpatialFilter(input_geometry)

    if filter:
        layer.SetAttributeFilter(filter)

    feature: ogr.Feature
    features: List[Dict] = []

    for feature in layer:
        json_str = feature.ExportToJson()
        features.append(json.loads(json_str))

    response = create_feature_collection(features, geojson_epsg)

    return response


@alru_cache(maxsize=32, ttl=_CACHE_TTL)
async def _get_geojson(url: Union[HttpUrl, FileUrl], timeout) -> str:
    if url.scheme == 'file':
        geojson = _load_geojson(url)
    else:
        _, geojson = await _fetch_geojson(url, timeout)

    return geojson


async def _fetch_geojson(url: HttpUrl, timeout) -> Tuple[int, str]:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                json_str = await response.text()

                return 200, json_str
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


def _load_geojson(file_uri: FileUrl) -> str:
    path = _file_uri_to_path(file_uri)

    try:
        with open(path) as file:
            return file.read()
    except:
        return None


def _file_uri_to_path(file_uri: FileUrl) -> str:
    parsed = urlparse(str(file_uri))

    return os.path.abspath(os.path.join(parsed.netloc, parsed.path))



from typing import List
from ...utils.helpers.common import parse_string
from ...models.config.quality_indicator import QualityIndicator


def get_threshold_values(quality_indicator: QualityIndicator) -> List[str]:
    if not quality_indicator.warning_threshold:
        return []
    
    values = [value.strip()
              for value in quality_indicator.warning_threshold.split('OR')]
    result = list(map(lambda value: parse_string(value), values))

    return result



from typing import List, Dict
from . import get_threshold_values
from ...models.quality_measurement import QualityMeasurement
from ...models.config.quality_indicator import QualityIndicator
from ...models.config.quality_indicator_type import QualityIndicatorType


def get_object_quality(quality_indicators: List[QualityIndicator], data: List[Dict]) -> tuple[List[QualityMeasurement], List[str]]:
    quality_data = _get_object_quality_data(quality_indicators, data)
    measurements: List[QualityMeasurement] = []
    warnings: List[str] = []

    for entry in quality_data:
        value: Dict

        for value in entry.get('values'):
            measurements.append(QualityMeasurement(entry.get('id'), entry.get(
                'name'), value.get('value'), value.get('comment')))

        warning = entry.get('warning_text')

        if warning is not None:
            warnings.append(warning)

    return measurements, warnings


def _get_object_quality_data(quality_indicators: List[QualityIndicator], data: List[Dict]) -> List[Dict]:
    if data is None or len(data) == 0:
        return []

    object_indicators = [
        indicator for indicator in quality_indicators if indicator.type == QualityIndicatorType.OBJECT]

    measurements: List[Dict] = []

    for oi in object_indicators:
        prop = oi.property
        threshold_values = get_threshold_values(oi)
        values: List[Dict] = []

        for entry in data:
            values.append({
                'value': entry[prop],
                'comment': None
            })

        distinct = list({value['value']: value for value in values}.values())

        should_warn = any(value['value'] for value in distinct if any(
            t_value for t_value in threshold_values if t_value == value['value']))

        measurements.append({
            'id': oi.quality_dimension_id,
            'name': oi.quality_dimension_name,
            'values': distinct,
            'warning_text': oi.quality_warning_text if should_warn else None
        })

    return measurements


__all__ = ['get_object_quality']


from uuid import UUID
from .analysis import Analysis
from .result_status import ResultStatus
from .config.dataset_config import DatasetConfig
from ..services.kartkatalog import get_kartkatalog_metadata


class EmptyAnalysis(Analysis):
    def __init__(self, config_id: UUID, config: DatasetConfig, result_status: ResultStatus):
        super().__init__(config_id, config, None, None, None, 0)
        self.result_status = result_status

    async def run(self):
        self.title = self.geolett['tittel'] if self.geolett else self.config.title
        self.themes = self.config.themes
        self.run_on_dataset = await get_kartkatalog_metadata(self.config.metadata_id)

    def _add_run_algorithm(self) -> None:
        raise NotImplementedError

    def _run_queries(self) -> None:
        return NotImplementedError

    def _set_distance_to_object(self) -> None:
        return NotImplementedError



import json
from io import BytesIO
from typing import List, Tuple, Dict, Union
from lxml import etree as ET
from osgeo import ogr
from ..models.config import CoverageService, CoverageGeoJson, CoverageGeoPackage
from ..drivers.wfs import query_wfs
from ..drivers.arcgis import query_arcgis
from ..drivers.geojson import query_geojson
from ..drivers.geopackage import query_geopackage
from ..utils.helpers.common import xpath_select_one, parse_string
from ..utils.helpers.geometry import geometry_from_gml, geometry_from_json


async def get_values_from_wfs(config: CoverageService, geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float, List[Dict]]:
    _, response = await query_wfs(config.url, config.layer, config.geom_field, geometry, epsg)

    if response is None:
        return [], 0, []

    source = BytesIO(response.encode('utf-8'))
    context = ET.iterparse(source, huge_tree=True)

    prop_path = f'.//*[local-name() = "{config.property}"]/text()'
    geom_path = f'.//*[local-name() = "{config.geom_field}"]/*'
    values: List[str] = []
    data: List[Dict] = []
    feature_geoms: List[ogr.Geometry] = []
    hit_area_percent = 0

    for _, elem in context:
        localname = ET.QName(elem).localname

        if localname == 'member':
            value = xpath_select_one(elem, prop_path)
            values.append(value)

            if value == 'ikkeKartlagt':
                geom_element = xpath_select_one(elem, geom_path)
                gml_str = ET.tostring(geom_element, encoding='unicode')
                feature_geom = geometry_from_gml(gml_str)

                if feature_geom:
                    feature_geoms.append(feature_geom)

            if len(config.properties) > 0:
                props = _map_wfs_properties(elem, config.properties)
                data.append(props)

    if len(feature_geoms) > 0:
        hit_area_percent = _get_hit_area_percent(geometry, feature_geoms)

    distinct_values = list(set(values))

    return distinct_values, hit_area_percent, data


async def get_values_from_arcgis(config: CoverageService, geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float, List[Dict]]:
    _, response = await query_arcgis(config.url, config.layer, None, geometry, epsg)

    if response is None:
        return [], 0, []

    features: List[Dict] = response.get('features')

    if len(features) == 0:
        return [], 0, []

    values: List[str] = []
    data: List[Dict] = []

    for feature in features:
        value = feature.get('properties').get(config.property)
        values.append(value)

        if len(config.properties) > 0:
            props = _map_geojson_properties(feature, config.properties)
            data.append(props)

    distinct_values = list(set(values))

    return distinct_values, 0, data


async def get_values_from_geojson(config: Union[CoverageGeoJson, CoverageGeoPackage], geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float, List[Dict]]:
    if isinstance(config, CoverageGeoJson):
        response = await query_geojson(config.url, config.filter, geometry, epsg)
    else:
        response = await query_geopackage(config.url, config.filter, geometry, epsg)

    if response is None:
        return [], 0, []

    features: List[Dict] = response.get('features')

    if len(features) == 0:
        return [], 0, []

    values: List[str] = []
    data: List[Dict] = []
    feature_geoms: List[ogr.Geometry] = []
    hit_area_percent = 0

    for feature in features:
        value = feature.get('properties').get(config.property)
        values.append(value)

        if value in ['ikkeKartlagt', 'Ikke kartlagt']:
            feature_geom = feature.get('geometry')
            json_str = json.dumps(feature_geom)
            geom = geometry_from_json(json_str)
            feature_geoms.append(geom)

        if len(config.properties) > 0:
            props = _map_geojson_properties(feature, config.properties)
            data.append(props)

    if len(feature_geoms) > 0:
        hit_area_percent = _get_hit_area_percent(geometry, feature_geoms)

    distinct_values = list(set(values))

    return distinct_values, hit_area_percent, data


def _get_hit_area_percent(geometry: ogr.Geometry, feature_geometries: List[ogr.Geometry]) -> float:
    geom_area: float = geometry.GetArea()
    hit_area: float = 0

    for geom in feature_geometries:
        intersection: ogr.Geometry = geom.Intersection(geometry)

        if intersection is None:
            continue

        area: float = intersection.GetArea()
        hit_area += area

    percent = (hit_area / geom_area) * 100

    return round(percent, 2)


def _map_wfs_properties(member: ET._Element, mappings: List[str]) -> Dict:
    props = {}

    for mapping in mappings:
        path = f'.//*[local-name() = "{mapping}"]/text()'
        value = xpath_select_one(member, path)

        if value:
            prop_name = mapping
            props[prop_name] = parse_string(value)

    return props


def _map_geojson_properties(feature: Dict, mappings: List[str]) -> Dict:
    props = {}
    feature_props: Dict = feature['properties']

    for mapping in mappings:
        props[mapping] = feature_props.get(mapping)

    return props


__all__ = ['get_values_from_wfs',
           'get_values_from_arcgis', 'get_values_from_geojson']



from typing import Optional
from pydantic import HttpUrl
from .coverage_base_service import CoverageBaseService


class CoverageService(CoverageBaseService):
    url: HttpUrl
    layer: str
    geom_field: Optional[str] = None



from .coverage_geojson import CoverageGeoJson
from .coverage_gpkg import CoverageGeoPackage
from .coverage_service import CoverageService
from .coverage_base_service import CoverageBaseService
from .dataset_config import DatasetConfig
from .layer import Layer
from .quality_config import QualityConfig
from .quality_indicator import QualityIndicator
from .quality_indicator_type import QualityIndicatorType


from os import getenv
from typing import Final
from ..utils.helpers.common import get_env_var

APP_FILES_DIR: Final[str] = get_env_var('APP_FILES_DIR')
CACHE_DIR:  Final[str] = f'{APP_FILES_DIR}/cache'
AR5_FGDB_PATH: Final[str] = getenv('AR5_FGDB_PATH')
SOCKET_IO_SRV_URL: Final[str] = getenv('SOCKET_IO_SRV_URL')
BLOB_STORAGE_CONN_STR: Final[str] = getenv('BLOB_STORAGE_CONN_STR')
MAP_IMAGE_API_URL: Final[str] = getenv('MAP_IMAGE_API_URL')
MAP_IMAGE_BASE_MAP: Final[str] = getenv('MAP_IMAGE_BASE_MAP') or 'WMTS'
DEFAULT_EPSG: Final[int] = 25833
WGS84_EPSG: Final[int] = 4326



from azure.storage.blob.aio import BlobServiceClient
from azure.storage.blob import PublicAccess, ContentSettings
from ..utils.constants import BLOB_STORAGE_CONN_STR


async def create_container(container_name: str) -> str:
    service = BlobServiceClient.from_connection_string(
        conn_str=BLOB_STORAGE_CONN_STR)

    try:
        container = await service.create_container(container_name, public_access=PublicAccess.BLOB)

        return container.url
    finally:
        await service.close()


async def upload_image(img: bytes, container_name: str, blob_name: str) -> str:
    service = BlobServiceClient.from_connection_string(
        conn_str=BLOB_STORAGE_CONN_STR)

    try:
        container_client = service.get_container_client(container_name)
        content_settings = ContentSettings(content_type='image/png')
        blob_client = await container_client.upload_blob(blob_name, img, content_settings=content_settings)

        return blob_client.url
    except:
        return None
    finally:
        await blob_client.close()
        await container_client.close()
        await service.close()


__all__ = ['create_container', 'upload_image']


from typing import List, Dict
from uuid import UUID
from . import get_threshold_values
from ..dok_status import get_dok_status_for_dataset
from ...utils.helpers.common import evaluate_condition
from ...models.quality_measurement import QualityMeasurement
from ...models.config import DatasetConfig
from ...models.config.quality_indicator import QualityIndicator
from ...models.config.quality_indicator_type import QualityIndicatorType


async def get_dataset_quality(config: DatasetConfig, quality_indicators: List[QualityIndicator], **kwargs) -> tuple[List[QualityMeasurement], List[str]]:
    quality_data = await _get_dataset_quality_data(config, quality_indicators, kwargs)
    measurements: List[QualityMeasurement] = []
    warnings: List[str] = []

    for entry in quality_data:
        value: Dict

        for value in entry.get('values'):
            measurements.append(QualityMeasurement(entry.get('id'), entry.get(
                'name'), value.get('value'), value.get('comment')))

        warning = entry.get('warning_text')

        if warning is not None:
            warnings.append(warning)

    return measurements, warnings


async def _get_dataset_quality_data(config: DatasetConfig, quality_indicators: List[QualityIndicator], data: Dict[str, any]) -> List[Dict]:
    quality_measurements = await _get_dataset_quality_measurements(config.metadata_id)

    dataset_indicators = [
        indicator for indicator in quality_indicators if indicator.type == QualityIndicatorType.DATASET]

    measurements: List[Dict] = []

    for qm in quality_measurements:
        id = qm['quality_dimension_id']
        name = qm['quality_dimension_name']
        value = qm['value']

        measurement = {
            'id': id,
            'name': name,
            'values': [{
                'value': value,
                'comment': qm['comment']
            }],
            'warning_text': None
        }

        di = next(
            (di for di in dataset_indicators if di.quality_dimension_id == id), None)

        if di is not None:
            measurement['warning_text'] = _get_dataset_quality_warning_text(
                value, di, data)

        measurements.append(measurement)

    return measurements


async def _get_dataset_quality_measurements(metadata_id: UUID) -> List[Dict]:
    qms: List[Dict] = []

    dok_status = await get_dok_status_for_dataset(metadata_id)

    if dok_status is not None:
        qms.extend(dok_status.get('suitability'))

    return qms


def _get_dataset_quality_warning_text(value: any, quality_indicator: QualityIndicator, data: Dict[str, any]) -> str:
    input_filter = quality_indicator.input_filter

    if input_filter is not None:
        result = evaluate_condition(input_filter, data)

        if not result:
            return None

    threshold_values = get_threshold_values(quality_indicator)
    should_warn = value in threshold_values

    return quality_indicator.quality_warning_text if should_warn else None


__all__ = ['get_dataset_quality']


from typing import List
from asyncio import Task, TaskGroup
from osgeo import ogr
from .area_types import get_area_types
from .buildings import get_buildings
from .roads import get_roads
from ...utils.helpers.geometry import create_buffered_geometry
from ...utils.helpers.map_image import create_payload_for_fact_sheet
from ...utils.constants import DEFAULT_EPSG, MAP_IMAGE_API_URL
from ...models.fact_sheet import FactSheet
from ...services.map_image import create_map_image


async def create_fact_sheet(geometry: ogr.Geometry, orig_epsg: int, buffer: int) -> FactSheet:
    fact_sheet = FactSheet()
    tasks = await _run_tasks(geometry, orig_epsg, buffer)

    for task in tasks:
        if task.get_name() == 'raster_result':
            fact_sheet.raster_result_image_bytes = task.result()
        else:
            result = task.result()

            if result:
                fact_sheet.fact_list.append(result)

    return fact_sheet


async def _run_tasks(geometry: ogr.Geometry, orig_epsg: int, buffer: int) -> List[Task]:
    input_geom = create_buffered_geometry(geometry, buffer, DEFAULT_EPSG)
    tasks: List[Task]

    async with TaskGroup() as tg:
        tasks = [
            tg.create_task(get_area_types(
                input_geom, DEFAULT_EPSG, orig_epsg, buffer)),
            tg.create_task(get_buildings(
                input_geom, DEFAULT_EPSG, orig_epsg, buffer)),
            # tg.create_task(get_roads(
            #     input_geom, DEFAULT_EPSG, orig_epsg, buffer)),
        ]

        if MAP_IMAGE_API_URL:
            tasks.append(tg.create_task(_create_raster_result(
                geometry, buffer), name='raster_result'))

    return tasks


async def _create_raster_result(geometry: ogr.Geometry, buffer: int):
    payload = create_payload_for_fact_sheet(geometry, buffer)
    _, image = await create_map_image(payload)

    return image


__all__ = ['create_fact_sheet']



import logging
from typing import Tuple, Dict
import aiohttp
import asyncio
from pydantic import HttpUrl
from osgeo import ogr
from ..utils.helpers.geometry import geometry_to_arcgis_geom

_LOGGER = logging.getLogger(__name__)


async def query_arcgis(url: HttpUrl, layer: str, filter: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Tuple[int, Dict]:
    api_url = f'{url}/{layer}/query'
    arcgis_geom = geometry_to_arcgis_geom(geometry, epsg)

    data = {
        'geometry': arcgis_geom,
        'geometryType': 'esriGeometryPolygon',
        'spatialRel': 'esriSpatialRelIntersects',
        'where': filter if filter is not None else '1=1',
        'inSR': epsg,
        'outSR': epsg,
        'units': 'esriSRUnit_Meter',
        'outFields': '*',
        'returnGeometry': True,
        'f': 'geojson'
    }

    return await _query_arcgis(api_url, data, timeout)


async def _query_arcgis(url: HttpUrl, data: Dict, timeout: int) -> Tuple[int, Dict]:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=data, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                json = await response.json()

                if 'error' in json:
                    return 400, None

                return 200, json
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


__all__ = ['query_arcgis']



import logging
from typing import Tuple, Dict
import aiohttp
import asyncio
from pydantic import HttpUrl
from osgeo import ogr
from ..utils.helpers.geometry import geometry_to_wkt
from ..utils.constants import WGS84_EPSG

_LOGGER = logging.getLogger(__name__)


async def query_ogc_api(base_url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, filter: str, epsg: int, out_epsg: int = 4326, timeout: int = 30) -> Tuple[int, Dict]:
    wkt_str = geometry_to_wkt(geometry, epsg)
    # autopep8: off
    filter_crs = f'&filter-crs=http://www.opengis.net/def/crs/EPSG/0/{epsg}' if epsg != WGS84_EPSG else ''
    crs = f'&crs=http://www.opengis.net/def/crs/EPSG/0/{out_epsg}' if out_epsg != WGS84_EPSG else ''
    url = f'{base_url}/{layer}/items?f=json&filter-lang=cql2-text{filter_crs}{crs}&filter=S_INTERSECTS({geom_field},{wkt_str})'

    if filter:
        url += f' AND {filter}'
    # autopep8: on

    return await _query_ogc_api(url, timeout)


async def _query_ogc_api(url: str, timeout: int) -> Tuple[int, Dict]:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                return 200, await response.json()
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


__all__ = ['query_ogc_api']



from typing import List, Dict
from osgeo import ogr
from .metadata import Metadata
from ..utils.helpers.geometry import create_run_on_input_geometry_json

class FactPart:
    def __init__(self, run_on_input_geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int, run_on_dataset: Metadata, run_algorithm: List[str], data: any):
        self.run_on_input_geometry = run_on_input_geometry
        self.epsg = epsg
        self.orig_epsg = orig_epsg
        self.buffer = buffer or 0
        self.run_on_dataset = run_on_dataset
        self.run_algorithm = run_algorithm
        self.data = data
        self.data_template: str = None

    def to_dict(self) -> Dict:
        return {
            'runOnInputGeometry': create_run_on_input_geometry_json(self.run_on_input_geometry, self.epsg, self.orig_epsg),
            'buffer': self.buffer,
            'runOnDataset': self.run_on_dataset.to_dict() if self.run_on_dataset is not None else None,
            'runAlgorithm': self.run_algorithm,
            'data': self.data,
            'dataTemplate': self.data_template
        }


import logging
import time
from uuid import UUID
from typing import Dict
from osgeo import ogr
from ..codelist import get_codelist
from ..kartkatalog import get_kartkatalog_metadata
from ...models.fact_part import FactPart
from ...utils.constants import AR5_FGDB_PATH

_LOGGER = logging.getLogger(__name__)

_METADATA_ID = UUID('166382b4-82d6-4ea9-a68e-6fd0c87bf788')
_LAYER_NAME = 'fkb_ar5_omrade'


async def get_area_types(geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int) -> FactPart:
    if not AR5_FGDB_PATH:
        return None

    start = time.time()
    dataset = await get_kartkatalog_metadata(_METADATA_ID)
    data = await _get_data(geometry)
    end = time.time()

    # autopep8: off
    _LOGGER.info(f'Fact sheet: Got area types from FKB AR5: {round(end - start, 2)} sec.')
    # autopep8: on

    return FactPart(geometry, epsg, orig_epsg, buffer, dataset, [f'intersect {_LAYER_NAME}'], data)


async def _get_data(geometry: ogr.Geometry) -> Dict:
    driver: ogr.Driver = ogr.GetDriverByName('OpenFileGDB')
    data_source: ogr.DataSource = driver.Open(AR5_FGDB_PATH, 0)
    layer: ogr.Layer = data_source.GetLayerByName(_LAYER_NAME)
    layer.SetSpatialFilter(0, geometry)

    input_area = geometry.GetArea()
    area_types = {}

    feature: ogr.Feature
    for feature in layer:
        area_type = feature.GetField('arealtype')
        geom: ogr.Geometry = feature.GetGeometryRef()
        intersection: ogr.Geometry = geometry.Intersection(geom)
        geom_area: float = intersection.GetArea()

        if area_type in area_types:
            area_types[area_type] += geom_area
        else:
            area_types[area_type] = geom_area

    return {
        'inputArea': round(input_area, 2),
        'areaTypes': await _map_area_types(area_types)
    }


async def _map_area_types(area_types: Dict) -> Dict:
    codelist = await get_codelist('arealressurs_arealtype')
    mapped = []

    for entry in codelist:
        label = entry['label']
        area: float = next((value for key, value in area_types.items()
                            if key == entry['value']), None)
        data = {'areaType': label}

        if area is not None:
            data['area'] = round(area, 2)
        else:
            data['area'] = 0.00

        mapped.append(data)

    return sorted(mapped, key=lambda item: item['areaType'])


__all__ = ['get_area_types']



from typing import List, Dict


class MapImagePayload:
    def __init__(self, width: int, height: int, base_map: Dict, wms: List[str], features: Dict, styling: Dict):
        self.width = width
        self.height = height
        self.base_map = base_map or None
        self.wms = wms or []
        self.features = features
        self.styling = styling or {}

    def to_dict(self) -> Dict:
        return {
            'width': self.width,
            'height': self.height,
            'baseMap': self.base_map,
            'wms': self.wms,
            'features': self.features,
            'styling': self.styling
        }



from typing import List, Dict
from osgeo import ogr
from .analysis import Analysis
from .fact_sheet import FactSheet
from ..utils.helpers.geometry import add_geojson_crs, create_buffered_geometry


class AnalysisResponse():
    result_list: List[Analysis]

    def __init__(self, input_geometry: Dict, input_geometry_area: float, fact_sheet: FactSheet, municipality_number: str, municipality_name: str):
        self.input_geometry = input_geometry
        self.input_geometry_area = input_geometry_area
        self.municipality_number = municipality_number
        self.municipality_name = municipality_name
        self.fact_sheet = fact_sheet or FactSheet()
        self.result_list = []
        self.report = None

    def to_dict(self) -> Dict:
        result_list = list(
            map(lambda analysis: analysis.to_dict(), self.result_list))
               
        data = {
            'resultList': result_list,
            'inputGeometry': self.input_geometry,
            'inputGeometryArea': self.input_geometry_area,
            'municipalityNumber': self.municipality_number,
            'municipalityName': self.municipality_name,
            'report': self.report
        }

        data = data | self.fact_sheet.to_dict()

        return data

    @classmethod
    def create(cls, geo_json: Dict, geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int, fact_sheet: FactSheet, municipality_number: str, municipality_name: str):
        add_geojson_crs(geo_json, orig_epsg)

        if buffer > 0:
            buffered_geom = create_buffered_geometry(geometry, buffer, epsg)
            geometry_area = round(buffered_geom.GetArea(), 2)
        else:
            geometry_area = round(geometry.GetArea(), 2)

        return AnalysisResponse(geo_json, geometry_area, fact_sheet, municipality_number, municipality_name)



from datetime import datetime
from typing import Self, Dict
from ..utils.helpers.common import parse_date_string


class Metadata:
    def __init__(self, dataset_id: str, title: str, description: str, owner: str, updated: datetime, dataset_description_uri: str):
        self.dataset_id = dataset_id
        self.title = title
        self.description = description
        self.owner = owner
        self.updated = updated
        self.dataset_description_uri = dataset_description_uri

    def to_dict(self) -> Dict:
        return {
            'datasetId': self.dataset_id,
            'title': self.title,
            'description': self.description,
            'owner': self.owner,
            'updated': datetime.isoformat(self.updated) if self.updated is not None else None,
            'datasetDescriptionUri': self.dataset_description_uri
        }

    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        dataset_id = data.get('datasetId')
        title = data.get('title')
        description = data.get('description')
        owner = data.get('owner')
        updated = data.get('updated')
        dataset_description_uri = data.get('datasetDescriptionUri')

        return Metadata(dataset_id, title, description, owner, parse_date_string(updated), dataset_description_uri)



from typing import List
from pydantic import BaseModel
import uuid


class CoverageBaseService(BaseModel):
    property: str
    planning_guidance_id: uuid.UUID = None
    building_guidance_id: uuid.UUID = None
    properties: List[str] = []    



import socketio
import logging
from .constants import SOCKET_IO_SRV_URL

_LOGGER = logging.getLogger(__name__)


def get_client() -> socketio.SimpleClient:
    if not SOCKET_IO_SRV_URL:
        return None

    try:
        sio = socketio.SimpleClient()
        sio.connect(SOCKET_IO_SRV_URL, socketio_path='/ws/socket.io')

        return sio
    except Exception as error:
        _LOGGER.warning(error)
        return None


__all__ = ['get_client']



import logging
import time
from uuid import UUID
from collections import Counter
from typing import List
from lxml import etree as ET
from osgeo import ogr
from ...drivers.wfs import query_wfs
from ...utils.helpers.common import parse_string
from ...services.kartkatalog import get_kartkatalog_metadata
from ...models.fact_part import FactPart

_LOGGER = logging.getLogger(__name__)

_METADATA_ID = UUID('24d7e9d1-87f6-45a0-b38e-3447f8d7f9a1')
_LAYER_NAME = 'Bygning'
_WFS_URL = 'https://wfs.geonorge.no/skwms1/wfs.matrikkelen-bygningspunkt'
_TIMEOUT = 10

_BUILDING_CATEGORIES = {
    (100, 159): 'Bolig',
    (160, 180): 'Fritidsbolig - hytte',
    (200, 299): 'Industri og lagerbygning',
    (300, 399): 'Kontor- og forretningsbygning',
    (400, 499): 'Samferdsels- og kommunikasjonsbygning',
    (500, 599): 'Hotell og restaurantbygning',
    (600, 699): 'Skole-, kultur-, idrett-, forskningsbygning',
    (700, 799): 'Helse- og omsorgsbygning',
    (800, 899): 'Fengsel, beredskapsbygning, mv.',
}


async def get_buildings(geometry: ogr.Geometry, epsg: int, orig_epsg: int, buffer: int) -> FactPart:
    start = time.time()
    dataset = await get_kartkatalog_metadata(_METADATA_ID)
    data = await _get_data(geometry, epsg)
    end = time.time()

    # autopep8: off
    _LOGGER.info(f'Fact sheet: Got buildings from Matrikkel WFS: {round(end - start, 2)} sec.')
    # autopep8: on

    return FactPart(geometry, epsg, orig_epsg, buffer, dataset, [f'intersect {_LAYER_NAME}'], data)


async def _get_data(geometry: ogr.Geometry, epsg: int) -> List[dict]:
    _, response = await query_wfs(_WFS_URL, _LAYER_NAME, 'representasjonspunkt', geometry, epsg, _TIMEOUT)

    if response is None:
        return None

    root = ET.fromstring(bytes(response, encoding='utf-8'))
    path = '//*[local-name() = "bygningstype"]'
    elems = root.xpath(path)
    categories = []

    for elem in elems:
        building_type = parse_string(elem.text)
        category = _get_building_category(building_type)

        if category is not None:
            categories.append(category)

    counted = Counter(categories)
    result: List[dict] = []

    for _, value in _BUILDING_CATEGORIES.items():
        count = counted.get(value, 0)

        result.append({
            'category': value,
            'count': count
        })

    return result


def _get_building_category(building_type: int) -> str:
    for range, category in _BUILDING_CATEGORIES.items():
        if range[0] <= building_type <= range[1]:
            return category

    return None


__all__ = ['get_buildings']



from typing import Union, Optional
from pydantic import FileUrl, HttpUrl
from .coverage_base_service import CoverageBaseService


class CoverageGeoPackage(CoverageBaseService):
    url: Union[FileUrl, HttpUrl]
    layer: str = '0'
    filter: Optional[str] = None


from pydantic import BaseModel
import uuid
from typing import Optional, List
from .quality_indicator import QualityIndicator


class QualityConfig(BaseModel):
    config_id: Optional[uuid.UUID] = None
    indicators: List[QualityIndicator]



from os import path
from pathlib import Path
import json
from typing import List, Dict
import aiohttp
from ..utils.constants import CACHE_DIR
from ..utils.helpers.common import should_refresh_cache

_CACHE_DAYS = 7

_CODELISTS = {
    'arealressurs_arealtype': 'https://register.geonorge.no/api/sosi-kodelister/fkb/ar5/5.0/arealressursarealtype.json',
    'fullstendighet_dekning': 'https://register.geonorge.no/api/sosi-kodelister/temadata/fullstendighetsdekningskart/dekningsstatus.json',
    'vegkategori': 'https://register.geonorge.no/api/sosi-kodelister/kartdata/vegkategori.json'
}


async def get_codelist(type: str) -> List[Dict] | None:
    url = _CODELISTS.get(type)

    if url is None:
        return None

    file_path = Path(
        path.join(CACHE_DIR, f'codelists/{type}.json'))

    if not file_path.exists() or should_refresh_cache(file_path, _CACHE_DAYS):
        file_path.parent.mkdir(parents=True, exist_ok=True)
        codelist = await _get_codelist(url)

        if codelist is None:
            return None

        json_object = json.dumps(codelist, indent=2)

        with file_path.open('w', encoding='utf-8') as file:
            file.write(json_object)

        return codelist
    else:
        with file_path.open(encoding='utf-8') as file:
            codelist = json.load(file)

        return codelist


async def _get_codelist(url: str) -> List[Dict]:
    response = await _fetch_codelist(url)

    if response is None:
        return None

    contained_items: List[Dict] = response.get('containeditems', [])
    entries: List[Dict] = []

    for item in contained_items:
        if item.get('status') == 'Gyldig':
            entries.append({
                'value': item.get('codevalue'),
                'label': item.get('label'),
                'description': item.get('description')
            })

    return entries


async def _fetch_codelist(url: str) -> Dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status != 200:
                    return None

                return await response.json()
    except:
        return None


__all__ = ['get_codelist']


from os import path
from uuid import UUID
from typing import Dict
import json
import aiohttp
from async_lru import alru_cache

_GEOLETT_API_URL = 'https://register.geonorge.no/geolett/api'
_LOCAL_GEOLETT_IDS = ['0c5dc043-e5b3-4349-8587-9b464d013aaa']
_CACHE_TTL = 86400 * 7


async def get_guidance_data(id: UUID) -> Dict:
    if id is None:
        return None

    if id in _LOCAL_GEOLETT_IDS:
        guidance_data = _fetch_local_guidance_data()
    else:
        guidance_data = await _fetch_guidance_data()

    result = list(filter(lambda item: item['id'] == str(id), guidance_data))

    return result[0] if len(result) > 0 else None


@alru_cache(maxsize=32, ttl=_CACHE_TTL)
async def _fetch_guidance_data() -> Dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(_GEOLETT_API_URL) as response:
                if response.status != 200:
                    return None

                return await response.json()
    except:
        return None


def _fetch_local_guidance_data() -> Dict:
    dir_path = path.dirname(path.realpath(__file__))

    file_path = path.join(
        path.dirname(dir_path), 'resources/geolett.local.json')

    with open(file_path, 'r') as file:
        return json.load(file)


__all__ = ['get_guidance_data']



from typing import List, Dict
from .fact_part import FactPart


class FactSheet:
    def __init__(self):
        self.raster_result_image: str = None
        self.raster_result_map: str = None
        self.raster_result_image: str = None
        self.raster_result_image_bytes: bytes = None
        self.cartography: str = None
        self.fact_list: List[FactPart] = []

    def to_dict(self) -> Dict:
        fact_list = list(
            map(lambda fact_part: fact_part.to_dict(), self.fact_list))
        
        return {
            'factSheetRasterResult': {
                'imageUri': self.raster_result_image,
                'mapUri': self.raster_result_map
            },
            'factSheetCartography': self.cartography,
            'factList': fact_list
        }



from typing import List, Dict, Tuple
from osgeo import ogr
from . import get_threshold_values
from ..coverage import get_values_from_wfs, get_values_from_arcgis, get_values_from_geojson
from ..codelist import get_codelist
from ...models.quality_measurement import QualityMeasurement
from ...models.config.quality_indicator import QualityIndicator


async def get_coverage_quality(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[List[QualityMeasurement], str, bool, List[Dict]]:
    quality_data, has_coverage, data = await _get_coverage_quality_data(quality_indicator, geometry, epsg)

    if quality_data is None:
        return [], None, False, []

    measurements: List[QualityMeasurement] = []

    for value in quality_data.get('values'):
        measurements.append(QualityMeasurement(quality_data.get('id'), quality_data.get(
            'name'), value.get('value'), value.get('comment')))

    warning = quality_data.get('warning_text')

    return measurements, warning, has_coverage, data


async def _get_coverage_quality_data(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[Dict[str, any], bool, List[Dict]]:
    values, hit_area_percent, data = await _get_values_from_web_service(quality_indicator, geometry, epsg)

    if len(values) == 0 and quality_indicator.warning_threshold is not None:
        return None, False, []

    codelist = await get_codelist('fullstendighet_dekning')
    meas_values: List[Dict] = []

    if len(values) == 0 and quality_indicator.warning_threshold is None:
        comment = _get_label_from_codelist('ikkeKartlagt', codelist)

        meas_values.append({
            'value': 'Nei',
            'comment': comment
        })
    else:
        for value in values:
            if value == 'Ikke kartlagt':
                value = 'ikkeKartlagt'
            elif value == 'Ikke relevant':
                value = 'ikkeRelevant'

            meas_value = 'Nei' if value in [
                'ikkeKartlagt', 'ikkeRelevant'] else 'Ja'

            comment = _get_label_from_codelist(value, codelist)

            meas_values.append({
                'value': meas_value,
                'comment': comment
            })

    measurement = {
        'id': quality_indicator.quality_dimension_id,
        'name': quality_indicator.quality_dimension_name,
        'values': meas_values,
        'warning_text': _get_warning_text(quality_indicator, values, hit_area_percent)
    }

    has_coverage = _has_coverage(values, quality_indicator.warning_threshold)

    return measurement, has_coverage, data


async def _get_values_from_web_service(quality_indicator: QualityIndicator, geometry: ogr.Geometry, epsg: int) -> Tuple[List[str], float, List[Dict]]:
    if quality_indicator.wfs:
        return await get_values_from_wfs(quality_indicator.wfs, geometry, epsg)

    if quality_indicator.arcgis:
        return await get_values_from_arcgis(quality_indicator.arcgis, geometry, epsg)

    if quality_indicator.geojson:
        return await get_values_from_geojson(quality_indicator.geojson, geometry, epsg)
    
    if quality_indicator.gpkg:
        return await get_values_from_geojson(quality_indicator.gpkg, geometry, epsg)

    # TODO: Add support for OGC Features API

    return [], 0, []


def _get_warning_text(quality_indicator: QualityIndicator, values: List[str], hit_area_percent: float) -> str:
    threshold_values = get_threshold_values(quality_indicator)

    should_warn = any(value for value in values if any(
        t_value for t_value in threshold_values if t_value == value))

    should_warn = should_warn or len(
        values) == 0 and quality_indicator.warning_threshold is None

    # not_relevant = 'ikkeRelevant' in values or 'Ikke relevant' in values
    # should_warn = should_warn or not_relevant
    warning_text = None

    if should_warn:
        warning_text: str = quality_indicator.quality_warning_text

        if 0 < hit_area_percent < 100:
            hit_area = str(hit_area_percent).replace('.', ',')
            warning_text = f'{hit_area} % av {warning_text.lower()}'

    return warning_text


def _has_coverage(values: List[str], warning_threshold: str) -> bool:
    if len(values) == 0 and warning_threshold is None:
        return False

    has_value = any(value in ['ikkeKartlagt', 'Ikke kartlagt',
                    'ikkeRelevant', 'Ikke relevant'] for value in values)

    if has_value:
        has_other_values = any(
            value not in ['ikkeKartlagt', 'Ikke kartlagt', 'ikkeRelevant', 'Ikke relevant'] for value in values)
        return has_other_values

    return True


def _get_label_from_codelist(value: str, codelist: List[Dict]) -> str:
    if codelist is None or len(codelist) == 0:
        return None

    result = next(
        (entry for entry in codelist if entry['value'] == value), None)

    return result.get('label') if result is not None else None


__all__ = ['get_coverage_quality']



from .analysis_response import AnalysisResponse
from .analysis import Analysis
from .arcgis_analysis import ArcGisAnalysis
from .empty_analysis import EmptyAnalysis
from .exceptions import DokAnalysisException
from .fact_part import FactPart
from .fact_sheet import FactSheet
from .map_image_payload import MapImagePayload
from .metadata import Metadata
from .ogc_api_analysis import OgcApiAnalysis
from .quality_measurement import QualityMeasurement
from .result_status import ResultStatus
from .wfs_analysis import WfsAnalysis


from osgeo import ogr
from .geometry import create_feature, create_feature_collection, create_buffered_geometry
from ..constants import DEFAULT_EPSG, MAP_IMAGE_BASE_MAP
from ...models.map_image_payload import MapImagePayload


def create_payload_for_analysis(geometry: ogr.Geometry, buffer: int, wms_url: str) -> MapImagePayload:
    base_map = {}

    if MAP_IMAGE_BASE_MAP == 'WMTS':
        base_map['wmts'] = {
            'url': 'https://cache.kartverket.no/v1/wmts/1.0.0/WMTSCapabilities.xml',
            'layer': 'topograatone'
        }
    else:
        base_map['osm'] = {
            'grayscale': True
        }

    features = [create_feature(geometry)]

    if buffer > 0:
        buffered_geometry = create_buffered_geometry(
            geometry, buffer, DEFAULT_EPSG)
        features.append(create_feature(buffered_geometry, {'buffer': True}))

    feature_collection = create_feature_collection(features, DEFAULT_EPSG)

    styling = [
        {
            'filter': ['has', 'buffer'],
            'style': {
                'stroke-color': '#d33333',
                'stroke-line-dash': [8, 8],
                'stroke-width': 2
            }
        },
        {
            'else': True,
            'style': {
                'stroke-color': '#d33333',
                'stroke-width': 4
            }
        }
    ]

    return MapImagePayload(1280, 720, base_map, [wms_url], feature_collection, styling)


def create_payload_for_fact_sheet(geometry: ogr.Geometry, buffer: int) -> MapImagePayload:
    base_map = {}

    if MAP_IMAGE_BASE_MAP == 'WMTS':
        base_map['wmts'] = {
            'url': 'https://cache.kartverket.no/v1/wmts/1.0.0/WMTSCapabilities.xml',
            'layer': 'topo'
        }
    else:
        base_map['osm'] = {
            'grayscale': False
        }

    features = [create_feature(geometry)]

    if buffer > 0:
        buffered_geometry = create_buffered_geometry(
            geometry, buffer, DEFAULT_EPSG)
        features.append(create_feature(buffered_geometry, {'buffer': True}))

    feature_collection = create_feature_collection(features, DEFAULT_EPSG)

    styling = [
        {
            'filter': ['has', 'buffer'],
            'style': {
                'stroke-color': '#d33333',
                'stroke-line-dash': [8, 8],
                'stroke-width': 2
            }
        },
        {
            'else': True,
            'style': {
                'stroke-color': '#d33333',
                'stroke-width': 4
            }
        }
    ]

    return MapImagePayload(1280, 548, base_map, None, feature_collection, styling)


__all__ = ['create_payload_for_analysis', 'create_payload_for_fact_sheet']


import logging
import os
import json
from urllib.parse import urlparse
from typing import List, Dict, Tuple, Union
from pydantic import HttpUrl, FileUrl
from pathlib import Path
from osgeo import ogr, osr
import asyncio
import aiohttp
import aiofiles
from ..utils.helpers.geometry import transform_geometry, create_feature_collection
from ..utils.helpers.common import should_refresh_cache
from ..utils.constants import APP_FILES_DIR

_LOGGER = logging.getLogger(__name__)
_CACHE_DAYS = 86400


async def query_geopackage(url: Union[HttpUrl, FileUrl], filter: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Dict:
    file_path = await _get_file_path(url, timeout)

    if not Path(file_path).exists():
        return None

    driver: ogr.Driver = ogr.GetDriverByName('GPKG')
    data_source: ogr.DataSource = driver.Open(file_path)
    layer: ogr.Layer = data_source.GetLayer(0)

    sr: osr.SpatialReference = layer.GetSpatialRef()
    auth_code: str = sr.GetAuthorityCode(None)
    gpkg_epsg = int(auth_code)

    if gpkg_epsg != epsg:
        input_geometry = transform_geometry(geometry, epsg, gpkg_epsg)
    else:
        input_geometry = geometry

    layer.SetSpatialFilter(input_geometry)

    if filter:
        layer.SetAttributeFilter(filter)

    feature: ogr.Feature
    features: List[Dict] = []

    for feature in layer:
        json_str = feature.ExportToJson()
        features.append(json.loads(json_str))

    response = create_feature_collection(features, gpkg_epsg)

    return response


async def _get_file_path(url: Union[HttpUrl, FileUrl], timeout: int) -> str:
    if url.scheme == 'file':
        return _file_uri_to_path(url)
    else:
        filename = _get_filename(url)
        file_path = Path(os.path.join(APP_FILES_DIR, f'geopackage/{filename}'))

        if not file_path.exists() or should_refresh_cache(file_path, _CACHE_DAYS):
            status, response = await _fetch_geopackage(url, timeout)

            if status != 200:
                return None

            file_path.parent.mkdir(parents=True, exist_ok=True)

            file = await aiofiles.open(file_path, mode='wb')
            await file.write(response)
            await file.close()

        return file_path.absolute()


async def _fetch_geopackage(url: HttpUrl, timeout) -> Tuple[int, bytes]:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                if response.status != 200:
                    return response.status, None

                return 200, await response.read()
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


def _get_filename(url: HttpUrl) -> str:
    parsed = urlparse(url)
    filename = os.path.basename(parsed.path)

    return filename.lower()


def _file_uri_to_path(file_uri: FileUrl) -> str:
    parsed = urlparse(str(file_uri))

    return os.path.abspath(os.path.join(parsed.netloc, parsed.path))



from os import path
import logging
from typing import Tuple
from io import BytesIO
from lxml import etree as ET
from pydantic import HttpUrl
import aiohttp
import asyncio
from osgeo import ogr

_LOGGER = logging.getLogger(__name__)


async def query_wfs(url: HttpUrl, layer: str, geom_field: str, geometry: ogr.Geometry, epsg: int, timeout: int = 30) -> Tuple[int, str]:
    gml_str = geometry.ExportToGML(['FORMAT=GML3'])
    request_xml = _create_wfs_request_xml(layer, geom_field, gml_str, epsg)

    return await _query_wfs(url, request_xml, timeout)


def _create_wfs_request_xml(layer: str, geom_field: str, gml_str: str, epsg: int) -> str:
    dir_path = path.dirname(path.realpath(__file__))
    file_path = path.join(dir_path, 'wfs_request.xml.txt')

    with open(file_path, 'r') as file:
        file_text = file.read()

    return file_text.format(layer=layer,  geom_field=geom_field, geometry=gml_str, epsg=epsg).encode('utf-8')


async def _query_wfs(url: HttpUrl, xml_body: str, timeout: int) -> Tuple[int, str]:
    url = f'{url}?service=WFS&version=2.0.0'
    headers = {'Content-Type': 'application/xml'}

    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=xml_body, headers=headers, timeout=timeout) as response:
                if response.status == 200:
                    return 200, await response.text()

                if response.status == 400:
                    _log_error_response(await response.text())

                return response.status, None
    except asyncio.TimeoutError:
        return 408, None
    except Exception as err:
        _LOGGER.error(err)
        return 500, None


def _log_error_response(response: str) -> None:
    source = BytesIO(response.encode('utf-8'))
    context = ET.iterparse(source)

    for _, elem in context:
        localname = ET.QName(elem).localname

        if localname != 'ExceptionText':
            continue

        _LOGGER.error(elem.text)


__all__ = ['query_wfs']
